/*
 * $Header: /home/puhuri/Tutkimus/src/timer/RCS/noworker.c,v 1.3 1997/02/19 12:46:44 puhuri Exp puhuri $
 *
 * $Log: noworker.c,v $
 * Revision 1.3  1997/02/19 12:46:44  puhuri
 * Ekojen mittausten versio.
 *
 * Revision 1.2  1996/11/22 08:25:26  puhuri
 * Compiles.  Muutettu uuteen rakenteeseen ja kyttmn conf-rakenneta.
 *
 * Revision 1.1  1996/11/14 12:07:51  puhuri
 * Initial revision
 *
 *
 *
 */
/* 
 * Includes source code from 'nsieve.c' by
 *          Sieve benchmark for variable sized arrays
 *                Version 1.2b, 26 Sep 1992
 *             Al Aburto (aburto@marlin.nosc.mil)
 *                      ('ala' on BIX)
 */

#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <math.h>
#include "sched_tst.h"

#ifdef RCS_INFO
static char rcs_id[] = "$Id: noworker.c,v 1.3 1997/02/19 12:46:44 puhuri Exp puhuri $";
#endif

static int noworker_child (struct conf_info *conf, int mynro);
static struct primes_dat sieve (char *ptr, long size);

int noworker (struct conf_info *conf)
{
  static int i = 0;
  pid_t pid;
  int try = 40;

  if (i >= conf->tprogs )	/* all started */
    return(0);

  conf->childs_running++;
  do {
    if ((pid = fork()) == 0) {	/* lapsiprosessi */
      noworker_child (conf, i);
      exit (OK);		/* palataan */
    }
  } while (pid < 0 && (errno == EAGAIN || errno == EINTR) && -- try > 0);

  if (pid < 0) {		/* fork eponnistui */
    conf->childs_running--;
    PERROR ("fork: ", ERR_FORK_FAILED | ERR_NOEXIT);
    return(1);
  }
  add_child_pid (conf, pid);
  i++;
  return (1);
}


struct primes_dat {
  size_t array_size;
  int limit;
  long found, last;
};

struct primes_dat Primes_dat[] = {
  /* Times on 180MHz PPro w/ Linux 2.0.18 856446515.766725 */
  { 2, 0, 3, 7 },		/* 856446515.766967 */
  { 4, 1, 4, 11 },		/* 856446515.766995 */
  { 8, 2, 7, 19 },		/* 856446515.767021 */
  { 16, 3, 10, 31 },		/* 856446515.767049 */
  { 32, 4, 18, 67 },		/* 856446515.767080 */
  { 64, 5, 31, 131 },		/* 856446515.767115 */
  { 128, 6, 54, 257 },		/* 856446515.767158 */
  { 256, 7, 96, 509 },		/* 856446515.767219 */
  { 512, 8, 171, 1021 },	/* 856446515.767315 */
  { 1024, 9, 308, 2039 },	/* 856446515.767479 */
  { 2048, 10, 564, 4099 },	/* 856446515.767777 */
  { 4096, 11, 1027, 8191 },	/* 856446515.768377 */
  { 8192, 12, 1899, 16381 },	/* 856446515.769570 */
  { 16384, 13, 3512, 32771 },	/* 856446515.772276 */
  { 32768, 14, 6543, 65539 },	/* 856446515.778212 */
  { 65536, 15, 12250, 131071 },	/* 856446515.793364 */
  { 131072, 16, 23000, 262147 }, /* 856446515.833735 */
  { 262144, 17, 43389, 524287 }, /* 856446515.970156 */
  { 524288, 18, 82024, 1048573 }, /* 856446516.289179 */
  { 1048576, 19, 155610, 2097143 }, /* 856446516.998068 */
  { 2097152, 20, 295946, 4194301 }, /* 856446518.593580 */
  { 4194304, 21, 564162, 8388593 }, /* 856446521.986611 */
  { 8388608, 22, 1077870, 16777213 }, /* 856446529.098399 */
  { 16777216, 23, 2063688, 33554393 }, /* 856446543.860214 */
  { 33554432, 24, 3957808, 67108859 }, /* 856446574.232809 */
  { 0, -1, 0, 0}
};

static int noworker_child (struct conf_info *conf, int mynro)
{
  struct timeval *tvp;
  int i;
  int class = 0;
  char *data_buf;

  if ((tvp = get_dirty_mem (conf->tloops * sizeof(struct timeval))) 
      == NULL) {
    exit (ERR_NO_MEMORY);
  }
  do {				/* find same size or larger */
    if (Primes_dat[class].array_size >= conf->data_size)
      break;
    class++;
  } while (Primes_dat[class].limit > 0);

  if (Primes_dat[class].limit == -1) {
    PERROR ("Sieve size: ", ERR_BAD_PARAMETER);
  }
  if ((data_buf = malloc (Primes_dat[class].array_size + 1)) == NULL) {
    exit (ERR_NO_MEMORY);
  }
  if (kill (getppid(), SIGUSR1)) {
    PERROR ("kill", ERR_KILL);
  }
  conf->wait_for_signal = 1;
  signal (SIGUSR1, child_wait_sig);

  while (conf->wait_for_signal)
    sleep (1);

  i = 0;
  for (;;) {
    struct primes_dat pd;

    if (gettimeofday (&(tvp[i]), NULL) < 0) {
      PERROR ("gettimeofday: ", ERR_GETTIMEOFDAY);
    }
    pd = sieve (data_buf, Primes_dat[class].array_size);

    if (pd.found != Primes_dat[class].found){
      printf ("sieve[%d]: Invalid value! [%d] found = %ld <> %ld\n",
	      i, class, pd.found, Primes_dat[class].found);
      exit (ERR_LOGIC);
    }
    if (pd.last != Primes_dat[class].last) {
      printf ("sieve[%d]: Invalid value! [%d] last = %ld <> %ld\n",
	      i, class, pd.last, Primes_dat[class].last);
      exit (ERR_LOGIC);
    }
    if ((tvp[i].tv_sec == conf->go_time.tv_sec && 
	 tvp[i].tv_usec < conf->go_time.tv_usec ) || 
	(tvp[i].tv_sec < conf->go_time.tv_sec )) {
      continue;
    }

    if (timercmp(&tvp[i], &(conf->stop_time), >))
      break;

    if (i < conf->tloops)
      i++;
  } 

  if (conf->usleep_end > 0)		/* maybe sleep some... */
#ifdef _HPUX_SOURCE		/* HPUX: ei usleep */
    sleep (conf->usleep_end / i10E6 ? : 1);
#else
    usleep (conf->usleep_end);
#endif

  printout_time_tval (conf, tvp, i, mynro, ID_CHAR_EMPTY);
  fprintf (stderr, 
	   "%d:%c mynro = %d\n",
	   getpid(), ID_CHAR_EMPTY, mynro);
  exit (OK);
}

/**************************************/
/*  Sieve of Erathosthenes Program    */
/**************************************/
#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

struct primes_dat sieve (char *ptr, long size)
{

  register char *flags;
  register long i, k;
  register long prime = 0;	/* assignment to silent gcc */
  register long count;
  struct primes_dat pd;

  flags = ptr;
  count = 0;                                       

  /* 
   * Intialize array
   */
  for(i=0 ; i<=size ; i++) {
    *(flags+i) = TRUE;	
  } 
  /* 
   * go thru array
   */
  for (i=0 ; i<=size ; i++) {                      
    if (*(flags+i)) {		/* if this a a prime */
      count++;			/* one more */
      prime = i + i + 3;	/* prime = 2*i + 3 */
      for(k = i + prime ; k<=size ; k+=prime){ 
	*(flags+k)=FALSE;	/* diviable by prime */
      } 
      
    }
  } 
  
  pd.array_size = size;
  pd.found = count;		/* number of primes */
  pd.last = prime;		/* last prime */

  return pd;
}                                               

